@@ -57,86 +57,129 @@ typedef struct _CdCache {
 #ifdef HAVE_HAL
   LibHalContext *ctx;
 #endif
+#ifdef HAVE_HAL_0_5
+  /* If the disc is a media, have the UDI available here */
+  char *disc_udi;
+#endif
 
   /* capabilities of the device */
   int cap;
 
+  /* Whether we have a medium */
+  guint has_medium : 1;
   /* if we're checking a media, or a dir */
-  gboolean is_media;
+  guint is_media : 1;
 
   /* indicates if we mounted this mountpoint ourselves or if it
    * was already mounted. */
-  gboolean self_mounted;
-  gboolean mounted;
+  guint self_mounted : 1;
+  guint mounted : 1;
 } CdCache;
 
 /* 
  * Resolve relative paths
  */
 
-static char *
-totem_disc_resolve_link (const char *dev, const char *buf)
+/* Copied from gtk+/gtk/gtkfilesystemunix.c */
+
+/* If this was a publically exported function, it should return
+ * a dup'ed result, but we make it modify-in-place for efficiency
+ * here, and because it works for us.
+ */
+static void
+canonicalize_filename (gchar *filename)
 {
-  char *parent, *new, *result;
+  gchar *p, *q;
+  gboolean last_was_slash = FALSE;
 
-  /* is it an absolute path? */
-  if (g_path_is_absolute (buf) != FALSE) {
-    result = realpath (buf, NULL);
-    if (result == NULL)
-      result = g_strdup (buf);
-  } else {
-    parent = g_path_get_dirname (dev);
-    new = g_build_filename (parent, buf, NULL);
-    result = realpath (new, NULL);
-    if (result == NULL)
-      result = g_strdup (new);
-    g_free (new);
-    g_free (parent);
-  }
+  p = filename;
+  q = filename;
 
-  return result;
+  while (*p)
+    {
+      if (*p == G_DIR_SEPARATOR)
+	{
+	  if (!last_was_slash)
+	    *q++ = G_DIR_SEPARATOR;
+
+	  last_was_slash = TRUE;
+	}
+      else
+	{
+	  if (last_was_slash && *p == '.')
+	    {
+	      if (*(p + 1) == G_DIR_SEPARATOR ||
+		  *(p + 1) == '\0')
+		{
+		  if (*(p + 1) == '\0')
+		    break;
+
+		  p += 1;
+		}
+	      else if (*(p + 1) == '.' &&
+		       (*(p + 2) == G_DIR_SEPARATOR ||
+			*(p + 2) == '\0'))
+		{
+		  if (q > filename + 1)
+		    {
+		      q--;
+		      while (q > filename + 1 &&
+			     *(q - 1) != G_DIR_SEPARATOR)
+			q--;
+		    }
+
+		  if (*(p + 2) == '\0')
+		    break;
+
+		  p += 2;
+		}
+	      else
+		{
+		  *q++ = *p;
+		  last_was_slash = FALSE;
+		}
+	    }
+	  else
+	    {
+	      *q++ = *p;
+	      last_was_slash = FALSE;
+	    }
+	}
+
+      p++;
+    }
+
+  if (q > filename + 1 && *(q - 1) == G_DIR_SEPARATOR)
+    q--;
+
+  *q = '\0';
 }
 
-/*
- * So, devices can be symlinks and that screws up.
- */
-
 static char *
-get_device (const char *device,
-	    GError     **error)
+totem_resolve_symlink (const char *device, GError **error)
 {
-  char *buf;
-  char *dev = g_strdup (device);
-  struct stat st;
+  char *dir, *link;
+  char *f;
+  char *f1;
 
-  while (1) {
-    char *new;
-
-    if (lstat (dev, &st) != 0) {
-      g_set_error (error, 0, 0,
-          _("Failed to find real device node for %s: %s"),
-          dev, g_strerror (errno));
-      g_free (dev);
+  f = g_strdup (device);
+  while (g_file_test (f, G_FILE_TEST_IS_SYMLINK)) {
+    link = g_file_read_link (f, error);
+    if(link == NULL) {
+      g_free (f);
       return NULL;
     }
 
-    if (!S_ISLNK (st.st_mode))
-      break;
-
-    if (!(buf = g_file_read_link (dev, NULL))) {
-      g_set_error (error, 0, 0,
-          _("Failed to read symbolic link %s: %s"),
-          dev, g_strerror (errno));
-      g_free (dev);
-      return NULL;
-    }
-    new = totem_disc_resolve_link (dev, buf);
-    g_free (dev);
-    g_free (buf);
-    dev = new;
+    dir = g_path_get_dirname (f);
+    f1 = g_build_filename (dir, link, NULL);
+    g_free (dir);
+    g_free (f);
+    f = f1;
   }
 
-  return dev;
+  if (f != NULL)
+    canonicalize_filename (f);
+  return f;
 }
 
 static gboolean
@@ -156,7 +199,8 @@ cd_cache_get_dev_from_volumes (GnomeVFSVolumeMonitor *mon, const char *device,
     volume = list->data;
     if (!(pdev = gnome_vfs_volume_get_device_path (volume)))
       continue;
-    if (!(pdev2 = get_device (pdev, NULL))) {
+    pdev2 = totem_resolve_symlink (pdev, NULL);
+    if (!pdev2) {
       g_free (pdev);
       continue;
     }
@@ -188,6 +232,7 @@ cd_cache_get_dev_from_volumes (GnomeVFSVolumeMonitor *mon, const char *device,
 
   return found;
 }
+
 static gboolean
 cd_cache_get_dev_from_drives (GnomeVFSVolumeMonitor *mon, const char *device,
 			      char **mountpoint, GnomeVFSDrive **d)
@@ -205,7 +250,8 @@ cd_cache_get_dev_from_drives (GnomeVFSVolumeMonitor *mon, const char *device,
     drive = list->data;
     if (!(pdev = gnome_vfs_drive_get_device_path (drive)))
       continue;
-    if (!(pdev2 = get_device (pdev, NULL))) {
+    pdev2 = totem_resolve_symlink (pdev, NULL);
+    if (!pdev2) {
       g_free (pdev);
       continue;
     }
@@ -217,7 +263,7 @@ cd_cache_get_dev_from_drives (GnomeVFSVolumeMonitor *mon, const char *device,
       mnt = gnome_vfs_drive_get_activation_uri (drive);
       if (mnt && strncmp (mnt, "file://", 7) == 0) {
         *mountpoint = g_strdup (mnt + 7);
-      } else /*if (mnt && strncmp (mnt, "cdda://", 7) == 0)*/ {
+      } else {
 	*mountpoint = NULL;
       }
       found = TRUE;
@@ -270,7 +316,8 @@ cd_cache_new_hal_ctx (void)
       libhal_ctx_free (ctx);
       return NULL;
     }
-    return ctx;
+    if (libhal_ctx_init (ctx, &error))
+      return ctx;
   }
 
   if (dbus_error_is_set (&error)) {
@@ -278,7 +325,6 @@ cd_cache_new_hal_ctx (void)
     dbus_error_free (&error);
   }
 
-  libhal_ctx_shutdown (ctx, NULL);
   libhal_ctx_free (ctx);
   if (conn != NULL)
     dbus_connection_unref (conn);
@@ -316,7 +362,7 @@ cd_cache_new (const char *dev,
   }
 
   /* retrieve mountpoint from gnome-vfs volumes and drives */
-  device = get_device (local, error);
+  device = totem_resolve_symlink (local, error);
   g_free (local);
   if (!device)
     return NULL;
@@ -408,11 +454,13 @@ cd_cache_has_medium (CdCache *cache)
       return FALSE;
     }
     retval = TRUE;
+    cache->disc_udi = udi;
+  } else {
+    g_free (udi);
   }
 
   if (devices != NULL)
     libhal_free_string_array (devices);
-  g_free (udi);
 
   return retval;
 }
@@ -449,7 +497,7 @@ cd_cache_open_device (CdCache *cache,
 		      GError **error)
 {
   /* not a medium? */
-  if (cache->is_media == FALSE) {
+  if (cache->is_media == FALSE || cache->has_medium != FALSE) {
     return TRUE;
   }
 
@@ -458,6 +506,7 @@ cd_cache_open_device (CdCache *cache,
 	_("Please check that a disc is present in the drive."));
     return FALSE;
   }
+  cache->has_medium = TRUE;
 
   return TRUE;
 }
@@ -560,7 +609,14 @@ cd_cache_free (CdCache *cache)
     conn = libhal_ctx_get_dbus_connection (cache->ctx);
     libhal_ctx_shutdown (cache->ctx, NULL);
     libhal_ctx_free(cache->ctx);
+#if (DBUS_VERSION_MAJOR == 0) && (DBUS_VERSION_MINOR < 60)
+    dbus_connection_unref (conn);
+#else
     dbus_connection_close (conn);
+    dbus_connection_unref (conn);
+#endif
+
+    g_free (cache->disc_udi);
 #elif HAVE_HAL_0_2
     hal_shutdown (cache->ctx);
 #endif
@@ -579,27 +635,55 @@ static MediaType
 cd_cache_disc_is_cdda (CdCache *cache,
 		       GError **error)
 {
-  MediaType type = MEDIA_TYPE_DATA;
-  GList *vol, *item;
+  MediaType type;
 
   /* We can't have audio CDs on disc, yet */
   if (cache->is_media == FALSE)
-    return type;
+    return MEDIA_TYPE_DATA;
+  if (!cd_cache_open_device (cache, error))
+    return MEDIA_TYPE_ERROR;
 
-  for (vol = item = gnome_vfs_drive_get_mounted_volumes (cache->drive);
-      item != NULL; item = item->next) {
-    char *mnt = gnome_vfs_volume_get_activation_uri (item->data);
-    if (mnt && strncmp (mnt, "cdda://", 7) == 0) {
-      g_free (mnt);
-      type = MEDIA_TYPE_CDDA;
-      break;
+#ifdef HAVE_HAL_0_5
+  {
+    DBusError error;
+    dbus_bool_t is_cdda;
+
+    dbus_error_init (&error);
+
+    is_cdda = libhal_device_get_property_bool (cache->ctx,
+	cache->disc_udi, "volume.disc.has_audio", &error);
+    type = is_cdda ? MEDIA_TYPE_CDDA : MEDIA_TYPE_DATA;
+
+    if (dbus_error_is_set (&error)) {
+      g_warning ("Error checking whether the volume is an audio CD: %s",
+	  error.message);
+      dbus_error_free (&error);
+      return MEDIA_TYPE_ERROR;
     }
-    g_free (mnt);
+    return type;
+  }
+#else
+  {
+    GList *vol, *item;
+
+    type = MEDIA_TYPE_DATA;
+
+    for (vol = item = gnome_vfs_drive_get_mounted_volumes (cache->drive);
+	item != NULL; item = item->next) {
+      char *mnt = gnome_vfs_volume_get_activation_uri (item->data);
+      if (mnt && strncmp (mnt, "cdda://", 7) == 0) {
+	g_free (mnt);
+	type = MEDIA_TYPE_CDDA;
+	break;
+      }
+      g_free (mnt);
+    }
+    g_list_foreach (vol, (GFunc) gnome_vfs_volume_unref, NULL);
+    g_list_free (vol);
   }
-  g_list_foreach (vol, (GFunc) gnome_vfs_volume_unref, NULL);
-  g_list_free (vol);
 
   return type;
+#endif
 }
 
 static gboolean
@@ -662,6 +746,36 @@ cd_cache_disc_is_vcd (CdCache *cache,
     return MEDIA_TYPE_ERROR;
   if (!cache->mountpoint)
     return MEDIA_TYPE_ERROR;
+#ifdef HAVE_HAL_0_5
+  if (cache->is_media != FALSE) {
+    DBusError error;
+    dbus_bool_t is_vcd;
+
+    dbus_error_init (&error);
+
+    is_vcd = libhal_device_get_property_bool (cache->ctx,
+	cache->disc_udi, "volume.disc.is_vcd", &error);
+
+    if (dbus_error_is_set (&error)) {
+      g_warning ("Error checking whether the volume is a VCD: %s",
+	  error.message);
+      dbus_error_free (&error);
+      return MEDIA_TYPE_ERROR;
+    }
+    if (is_vcd != FALSE)
+      return MEDIA_TYPE_VCD;
+    is_vcd = libhal_device_get_property_bool (cache->ctx,
+	cache->disc_udi, "volume.disc.is_svcd", &error);
+
+    if (dbus_error_is_set (&error)) {
+      g_warning ("Error checking whether the volume is an SVCD: %s",
+	  error.message);
+      dbus_error_free (&error);
+      return MEDIA_TYPE_ERROR;
+    }
+    return is_vcd ? MEDIA_TYPE_VCD : MEDIA_TYPE_DATA;
+  }
+#endif
   /* first is VCD, second is SVCD */
   if (cd_cache_file_exists (cache, "MPEGAV", "AVSEQ01.DAT") ||
       cd_cache_file_exists (cache, "MPEG2", "AVSEQ01.MPG"))
@@ -681,6 +795,25 @@ cd_cache_disc_is_dvd (CdCache *cache,
     return MEDIA_TYPE_ERROR;
   if (!cache->mountpoint)
     return MEDIA_TYPE_ERROR;
+#ifdef HAVE_HAL_0_5
+  if (cache->is_media != FALSE) {
+    DBusError error;
+    dbus_bool_t is_dvd;
+
+    dbus_error_init (&error);
+
+    is_dvd = libhal_device_get_property_bool (cache->ctx,
+	cache->disc_udi, "volume.disc.is_videodvd", &error);
+
+    if (dbus_error_is_set (&error)) {
+      g_warning ("Error checking whether the volume is a DVD: %s",
+	  error.message);
+      dbus_error_free (&error);
+      return MEDIA_TYPE_ERROR;
+    }
+    return is_dvd ? MEDIA_TYPE_DVD : MEDIA_TYPE_DATA;
+  }
+#endif
   if (cd_cache_file_exists (cache, "VIDEO_TS", "VIDEO_TS.IFO"))
     return MEDIA_TYPE_DVD;
 
